﻿/*
	CVH.cs - Allegiance CVH files API 
	Copyright (C) Kirth Gersen, 2001-2008.  All rights reserved.
	v 1.0 - SlimDX version
*/

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using SlimDX;
using System.Text;
using System.Globalization;
using QuickHull3D;

namespace Allegiance
{
    public struct CVHFace
    {
        public short I1;
        public short I2;
        public short I3;
        short minI;
        short midI;
        short maxI;
        public CVHFace(short i1, short i2, short i3)
        {
            I1 = i1; I2 = i2; I3 = i3;
            minI = Math.Min(I1, Math.Min(I2, I3));
            maxI = Math.Max(I1, Math.Max(I2, I3));
            midI = 0;
            if ((minI == I1) && (maxI == I2)) midI = I3;
            if ((minI == I2) && (maxI == I1)) midI = I3;

            if ((minI == I2) && (maxI == I3)) midI = I1;
            if ((minI == I3) && (maxI == I2)) midI = I1;

            if ((minI == I1) && (maxI == I3)) midI = I2;
            if ((minI == I3) && (maxI == I1)) midI = I2;

        }
        public static bool operator ==(CVHFace left, CVHFace right)
        {
            return (left.minI == right.minI) && (left.maxI == right.maxI) && (left.midI == right.midI);
        }
        public static bool operator !=(CVHFace left, CVHFace right)
        {
            return !(left == right);
        }
        public override int GetHashCode()
        {
            // todo : provide a unique hashcode given I1,I2,I3 - but it's not quite possible
            // unless we limit the vertices count
            return base.GetHashCode();
        }
        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }
    }
    public class CVHMesh : HashSet<CVHFace>
    {
        public Vector3[] Vertices { get; set; }
        public short[] Faces {get ; set;}
    }
    public struct ConvexVertex
    {
        public Vector3 position;
        public int[] Adjacencies;
    }
    public struct ConvexHull
    {
        public Vector3 Center;
        public int nVertices;
        public int nAdjacencies;
        public ConvexVertex[] Vertices;
        // generate an indices buffer for faces (triangles)
        public CVHMesh Mesh { get; set; }
        public void GenMesh()
        {
            Point3d[] points = new Point3d[Vertices.Length];
            int i = 0;
            foreach (ConvexVertex vert in Vertices)
            {
                points[i++] = new Point3d(vert.position.X, vert.position.Y, vert.position.Z);
            }
            QuickHull3D.QuickHull3D qh = new QuickHull3D.QuickHull3D(points);
            qh.triangulate();
            points = qh.getVertices();
            Mesh = new CVHMesh();
            Mesh.Vertices = new Vector3[points.Length];
            i = 0;
            foreach (Point3d point in points)
            {
                Mesh.Vertices[i++] = new Vector3((float)point.x, (float)point.y, (float)point.z);
            }
            int[][] faces = qh.getFaces();
            Mesh.Faces = new short[qh.getNumFaces()*3];
            i = 0;
            foreach (int[] face in faces)
            {
                Mesh.Faces[i++] = (short)face[0];
                Mesh.Faces[i++] = (short)face[1];
                Mesh.Faces[i++] = (short)face[2];
            }
        }
    }
    public class CVHFrame
    {
        public string Name;
        public Vector3 Position;
        public Vector3 Forward;
    }
    public class CVH
    {

        public const int c_htsConvexHullMin = 0;
        public const int c_htsConvexHullMax = 127;

        public float radius;
        public Vector3 EllipseEquation;
        public float erm;
        public int nHulls;
        public ConvexHull[] Hulls;
        public List<CVHFrame> Frames;

        public CVH()
        {
            Frames = new List<CVHFrame>();
        }
        public void AddFrame(string name, Vector3 position, Vector3 forward)
        {
            CVHFrame frame = new CVHFrame();
            frame.Name = name;
            frame.Position = position;
            frame.Forward = forward;
            Frames.Add(frame);
        }
        public static CVH FromFile(string filename)
        {
            CVH cvh = new CVH();
            FileStream cf;
            cf = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.Read);
            StreamReader sr = new StreamReader(cf);
            string line = sr.ReadLine();
            string[] split = line.Split(new char[] {' '},StringSplitOptions.RemoveEmptyEntries);
            cvh.radius = float.Parse(split[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
            float X = float.Parse(split[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
            float Y = float.Parse(split[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
            float Z = float.Parse(split[3], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
            cvh.EllipseEquation = new Vector3(X, Y, Z);
            cvh.erm = float.Parse(split[4], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);

            line = sr.ReadLine();
            int nHulls = int.Parse(line);
            if (nHulls >= c_htsConvexHullMax) throw new ArgumentException("too many hulls");

            cvh.nHulls = nHulls;
            cvh.Hulls = new ConvexHull[nHulls];

            for (int h = 0; h < nHulls; h++)
            {

                int     nVertices;
                int     nAdjacencies;
                Vector3  center;
                line = sr.ReadLine(); split = line.Split(new char[] {' '},StringSplitOptions.RemoveEmptyEntries);
                nVertices = int.Parse(split[0]);
                nAdjacencies = int.Parse(split[1]);
                X = float.Parse(split[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Y = float.Parse(split[3], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Z = float.Parse(split[4], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);

                // hack LHS
                center = new Vector3(X,Y,-Z);

                cvh.Hulls[h].Center = center;
                cvh.Hulls[h].nVertices = nVertices;
                cvh.Hulls[h].nAdjacencies = nAdjacencies;
                cvh.Hulls[h].Vertices = new ConvexVertex[nVertices];

                for (int v = 0; v < nVertices; v++)
                {
                    line = sr.ReadLine();  split = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                    X = float.Parse(split[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                    Y = float.Parse(split[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                    Z = float.Parse(split[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                    int n = int.Parse(split[3]);

                    // hack LHS
                    Vector3 position = new Vector3(X, Y, -Z);
                    cvh.Hulls[h].Vertices[v] = new ConvexVertex();
                    cvh.Hulls[h].Vertices[v].position = position;
                    cvh.Hulls[h].Vertices[v].Adjacencies = new int[n];
                    line = sr.ReadLine(); split = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                    for (int a = 0; a < n; a++)
                        cvh.Hulls[h].Vertices[v].Adjacencies[a] = int.Parse(split[a]);
                }
            }
            // parse frame data
            do
            {
                try { line = sr.ReadLine(); }
                catch (EndOfStreamException) 
                {
                    break;
                }
                if (line == null) break;
                string name = line.Trim();
                if (name == "") break;

                line = sr.ReadLine(); split = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                X = float.Parse(split[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Y = float.Parse(split[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Z = float.Parse(split[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Vector3 position = new Vector3(-X, Y, Z);

                line = sr.ReadLine(); split = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                X = float.Parse(split[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Y = float.Parse(split[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Z = float.Parse(split[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                Vector3 forward = new Vector3(X, -Y, -Z);

                cvh.AddFrame(name, position, forward);

            } while (true);
            sr.Close();
            cf.Close();
            return cvh;
        }
        public void ToFile(string filename)
        {
            throw new NotImplementedException();
        }
    }
}
