//
//      coded by un
//            --------------
//                     mindshifter.com
//


using System;
using System.Collections.Generic;
using SlimDX;
using SlimDX.Direct3D9;
using Microsoft.Samples.DirectX.UtilityToolkit;
using System.Drawing;

namespace Mindshifter
{
    // Describes a vertex transformation stage matrix
    public enum MatrixMode
    {
        World,
        View,
        Projection
    }

    // internal delegate to handle queued renderables
    internal delegate void PickRenderer();

    // public delegate to support custom draw calls for pick buffer clients
    public delegate void OverridePickRenderer(Device device);

    internal struct PickRenderable
    {
        public PickRenderer    Render;
        public Matrix          WVP;
        public Color4          PickColor;

        internal PickRenderable(PickRenderer render, Matrix wvp, Color4 pick_color)
        {
            Render = render;
            WVP = wvp;
            PickColor = pick_color;
        }
    }

    public class PickBuffer
    {
        private Framework                   mFramework;

        private Surface             mBuffer;//RenderTarget2D
        private Surface             mPickSurface;//Texture2D
        private Surface             mDepthBuffer;//DepthStencilBuffer

        private PickEffect                  mEffect;

        private Queue<PickRenderable>       mRenderables;

        private Dictionary<MatrixMode,
                        Stack<Matrix>>      mMatrices;

        private Stack<VertexDeclaration>    mVertexDeclaration;
        private Stack<VertexBuffer>         mVertexBuffer;
        private Stack<IndexBuffer>          mIndexBuffer;

        private Stack<uint>                 mPickID;

        /// <summary>
        /// Gets the size of the render queue for the current frame
        /// </summary>
        public int QueueSize
        {
            get { return mRenderables.Count; }
        }

        /// <summary>
        /// Gets the size of the view matrix stack
        /// </summary>
        public int ViewMatrixStackSize
        {
            get { return mMatrices[MatrixMode.View].Count; }
        }

        /// <summary>
        /// Gets the size of the projection matrix stack
        /// </summary>
        public int ProjectionMatrixStackSize
        {
            get { return mMatrices[MatrixMode.Projection].Count; }
        }

        /// <summary>
        /// Gets the size of the world matrix stack
        /// </summary>
        public int WorldMatrixStackSize
        {
            get { return mMatrices[MatrixMode.World].Count; }
        }

        /// <summary>
        /// Gets the size of the pick id stack
        /// </summary>
        public int PickIDStackSize
        {
            get { return mPickID.Count; }
        }

        /// <summary>
        /// Gets the size of the vertex buffer stack
        /// </summary>
        public int VertexBufferStackSize
        {
            get { return mVertexBuffer.Count; }
        }

        /// <summary>
        /// Gets the size of the index buffer stack
        /// </summary>
        public int IndexBufferStackSize
        {
            get { return mIndexBuffer.Count; }
        }

        /// <summary>
        /// Gets the size of the vertex declaration stack
        /// </summary>
        public int VertexDeclarationStackSize
        {
            get { return mVertexDeclaration.Count; }
        }

        /// <summary>
        /// Creates a new instance of PickBuffer
        /// </summary>
        /// <param name="graphics">The IGraphicsDeviceService to be used for rendering</param>
        public PickBuffer(Framework frm)
        {
            // store graphics device service
            mFramework = frm;

            // Render queue
            mRenderables = new Queue<PickRenderable>();

            // Matrix stacks for each MatrixMode
            mMatrices = new Dictionary<MatrixMode, Stack<Matrix>>();
            for (MatrixMode m = MatrixMode.World; m <= MatrixMode.Projection; ++m)
                mMatrices[m] = new Stack<Matrix>();

            // Vertex and index data stacks
            mVertexDeclaration = new Stack<VertexDeclaration>();
            mVertexBuffer = new Stack<VertexBuffer>();
            mIndexBuffer = new Stack<IndexBuffer>();

            // Pick id stack
            mPickID = new Stack<uint>();

            //
            mBuffer = null;
            mDepthBuffer = null;
            // Hook into the device created and reset events so we can re-create our
            // render target buffer when necessary
            mFramework.DeviceCreated += new DeviceEventHandler(CreateBuffer);
            mFramework.DeviceReset += new DeviceEventHandler(CreateBuffer);
            mFramework.DeviceLost += new EventHandler(DisposeBuffer);
            mFramework.Disposing += new EventHandler(DisposeBuffer);

            // If the graphics device is already up and running we won't hit the above events,
            // so check manually for whether or not we need to create the render target
            // buffer right now
            if ((mFramework.Device != null) && !mFramework.Device.Disposed)
                CreateBuffer(this, null);

            mEffect = new PickEffect(mFramework);
        }

        /// <summary>
        /// Creates the render target used to store pick id data
        /// </summary>
        /// <param name="sender">unused</param>
        /// <param name="e">unused</param>
        private void CreateBuffer(object sender, EventArgs e)
        {
            DisposeBuffer(sender,e);

            Device dev = mFramework.Device;

            SurfaceDescription prm = mFramework.BackBufferSurfaceDescription;

            int bbw = prm.Width;
            int bbh = prm.Height;

            mBuffer = Surface.CreateRenderTarget(dev, bbw, bbh, Format.A8R8G8B8, MultisampleType.None, 0, false);
            mPickSurface = Surface.CreateOffscreenPlain(dev, bbw, bbh, Format.A8R8G8B8, Pool.SystemMemory);

            Surface dss = dev.DepthStencilSurface;

            mDepthBuffer = Surface.CreateDepthStencil(dev, bbw, bbh, dss.Description.Format, MultisampleType.None, 0, false);
                //new DepthStencilBuffer(dev, bbw, bbh, dev.DepthStencilBuffer.Format);
            dss.Dispose();
        }
        private void DisposeBuffer(object sender, EventArgs e)
        {
            if (mBuffer != null) 
                if (!mBuffer.Disposed)
                    mBuffer.Dispose();
            mBuffer = null;
            if (mDepthBuffer != null)
                if (!mDepthBuffer.Disposed)
                    mDepthBuffer.Dispose();
            mDepthBuffer = null;
            if (mPickSurface != null)
                if (!mPickSurface.Disposed)
                    mPickSurface.Dispose();
            mPickSurface = null;
        }
        /// <summary>
        /// Pushes a matrix of the specified MatrixMode onto the matrix stack
        /// </summary>
        /// <param name="mode">The type of matrix to push onto the corresponding stack</param>
        /// <param name="mtx">The matrix to push onto the stack</param>
        public void PushMatrix(MatrixMode mode, Matrix mtx)
        {
            mMatrices[mode].Push(mtx);
        }

        /// <summary>
        /// Pops a matrix of the specified MatrixMode from the matrix stack
        /// </summary>
        /// <param name="mode">The type of matrix to pop from the corresponding stack</param>
        /// <returns>The matrix popped from the stack</returns>
        public Matrix PopMatrix(MatrixMode mode)
        {
            if (mMatrices[mode].Count == 0)
                throw new InvalidOperationException(string.Format("Attempt to pop a {0} Matrix from a stack of size 0", mode.ToString()));

            return mMatrices[mode].Pop();
        }

        /// <summary>
        /// Gets the topmost matrix from the specified MatrixMode stack
        /// </summary>
        /// <param name="mode">The type of matrix to retrieve</param>
        /// <returns>The topmost matrix on the specified MatrixMode stack</returns>
        public Matrix GetMatrix(MatrixMode mode)
        {
            if (mMatrices[mode].Count == 0)
                throw new InvalidOperationException(string.Format("Attempt to retrieve a {0} Matrix from a stack of size 0", mode.ToString()));

            return mMatrices[mode].Peek();
        }

        /// <summary>
        /// Pushes a VertexDeclaration onto the vertex declaration stack
        /// </summary>
        /// <param name="decl">The vertex declaration to push onto the stack</param>
        public void PushVertexDeclaration(VertexDeclaration decl)
        {
            mVertexDeclaration.Push(decl);
        }

        public VertexDeclaration PopVertexDeclaration()
        {
            if (mVertexDeclaration.Count == 0)
                throw new InvalidOperationException("Attempt to pop a VertexDeclaration from a stack of size 0");

            return mVertexDeclaration.Pop();
        }

        public VertexDeclaration GetVertexDeclaration()
        {
            if (mVertexDeclaration.Count == 0)
                throw new InvalidOperationException("Attempt to retrieve a VertexDeclaration from a stack of size 0");

            return mVertexDeclaration.Peek();
        }

        public void PushVertexBuffer(VertexBuffer vb)
        {
            mVertexBuffer.Push(vb);
        }

        public VertexBuffer PopVertexBuffer()
        {
            if(mVertexBuffer.Count == 0)
                throw new InvalidOperationException("Attempt to pop a VertexBuffer from a stack of size 0");

            return mVertexBuffer.Pop();
        }

        public VertexBuffer GetVertexBuffer()
        {
            if (mVertexBuffer.Count == 0)
                throw new InvalidOperationException("Attempt to retrieve a VertexBuffer from a stack of size 0");

            return mVertexBuffer.Peek();
        }

        public void PushIndexBuffer(IndexBuffer ib)
        {
            mIndexBuffer.Push(ib);
        }

        public IndexBuffer PopIndexBuffer()
        {
            if(mIndexBuffer.Count == 0)
                throw new InvalidOperationException("Attempt to pop an IndexBuffer from a stack of size 0");

            return mIndexBuffer.Pop();
        }

        public IndexBuffer GetIndexBuffer()
        {
            if (mIndexBuffer.Count == 0)
                throw new InvalidOperationException("Attempt to retrieve a VertexBuffer from a stack of size 0");

            return mIndexBuffer.Peek();
        }

        public void PushPickID(uint id)
        {
            mPickID.Push(id);
        }

        public uint PopPickID()
        {
            if (mPickID.Count == 0)
                throw new InvalidOperationException("Attempt to pop a pick id from a stack of size 0");

            return mPickID.Pop();
        }

        public uint GetPickID()
        {
            if (mPickID.Count == 0)
                throw new InvalidOperationException("Attempt to retrieve a pick id from a stack of size 0");

            return mPickID.Peek();
        }

        private Color4 CalculatePickColor(uint pick_id)
        {
            byte r = (byte)((pick_id >> 24) & 0xFF);
            byte g = (byte)((pick_id >> 16) & 0xFF);
            byte b = (byte)((pick_id >> 8) & 0xFF);
            byte a = (byte)(pick_id & 0xFF);

            Color4 col = new Color4(Color.FromArgb(a, r, g, b));
            return col;
        }

        private uint CalculatePickID(uint color)
        {
            uint id = 0;

            id |= ((uint)(color & 0x00FFFFFF) << 8);
            id |= ((uint)(color & 0xFF000000) >> 24);
            return id;
        }

        private void QueueRenderable(PickRenderer handler)
        {
            Matrix view = GetMatrix(MatrixMode.View);
            Matrix proj = GetMatrix(MatrixMode.Projection);
            Matrix world = GetMatrix(MatrixMode.World);
            uint pick_id = GetPickID();

            Color4 col = CalculatePickColor(pick_id);
            
            Matrix wvp = world * view * proj;

            mRenderables.Enqueue(new PickRenderable(handler, wvp, col));
        }

        //public void QueuePrimitives(PrimitiveType prim_type, int buffer_ofst, int start_vtx, int prim_count)
        //{
        //    VertexDeclaration decl = GetVertexDeclaration();
        //    VertexBuffer vtx_buf = GetVertexBuffer();
        //    int stride = D3DX.GetFVFVertexSize(vtx_buf.Description.FVF); 

        //    QueueRenderable(delegate()
        //    {
        //        mFramework.Device.VertexDeclaration = decl;
        //        mFramework.Device.SetStreamSource(0, vtx_buf, buffer_ofst, stride);
        //        mFramework.Device.DrawPrimitives(prim_type, start_vtx, prim_count);
        //    });
        //}

        //public void QueueIndexedPrimitives(PrimitiveType prim_type, int buffer_ofst, int base_vtx, int min_vtx_idx,
        //    int num_vtx, int start_idx, int prim_count)
        //{
        //    Device dev = mFramework.Device;

        //    VertexDeclaration decl = GetVertexDeclaration();
        //    VertexBuffer vtx_buf = GetVertexBuffer();
        //    IndexBuffer idx_buf = GetIndexBuffer();
        //    int stride = D3DX.GetFVFVertexSize(vtx_buf.Description.FVF);

        //    QueueRenderable(delegate()
        //    {
        //        dev.VertexDeclaration = decl;
        //        dev.SetStreamSource(0,vtx_buf, buffer_ofst, stride);
        //        dev.Indices = idx_buf;
        //        dev.DrawIndexedPrimitives(prim_type, base_vtx, min_vtx_idx,
        //            num_vtx, start_idx, prim_count);
        //    });
        //}

        //public void QueueUserPrimitives<T>(PrimitiveType prim_type, T[] vtx_data, int vtx_ofst, int prim_count) 
        //    where T : struct
        //{
        //    Device dev = mFramework.Device;

        //    VertexDeclaration decl = GetVertexDeclaration();

        //    QueueRenderable(delegate()
        //    {
        //        dev.VertexDeclaration = decl;
        //        dev.DrawUserPrimitives<T>(prim_type, vtx_ofst, prim_count, vtx_data);
        //    });
        //}

        //public void QueueUserIndexedPrimitives<T>(PrimitiveType prim_type, T[] vtx_data, int vtx_ofst,
        //    int num_vtx, int[] idx_data, int idx_ofst, int prim_count) where T : struct
        //{
        //    Device dev = mFramework.Device;

        //    VertexDeclaration decl = GetVertexDeclaration();

        //    QueueRenderable(delegate()
        //    {
        //        dev.VertexDeclaration = decl;
        //        dev.DrawIndexedUserPrimitives<int,T>(prim_type,vtx_ofst,num_vtx,prim_count,idx_data,Format.Index32,vtx_data,sizeof(T));
        //            //prim_type, vtx_data, vtx_ofst, num_vtx,
        //            //idx_data, idx_ofst, prim_count);
        //    });
        //}

        //public void QueueUserIndexedPrimitives<T>(PrimitiveType prim_type, T[] vtx_data, int vtx_ofst,
        //    int num_vtx, short[] idx_data, int idx_ofst, int prim_count) where T : struct
        //{
        //    Device dev = mFramework.Device;

        //    VertexDeclaration decl = GetVertexDeclaration();

        //    QueueRenderable(delegate()
        //    {
        //        dev.VertexDeclaration = decl;
        //        dev.DrawUserIndexedPrimitives(prim_type, vtx_data, vtx_ofst, num_vtx,
        //            idx_data, idx_ofst, prim_count);
        //    });
        //}

        public void QueueRenderFunction(OverridePickRenderer renderer)
        {
            Device dev = mFramework.Device;

            VertexDeclaration decl = GetVertexDeclaration();

            QueueRenderable(delegate()
            {
                dev.VertexDeclaration = decl;
                renderer(mFramework.Device);
            });
        }

        public void Render()
        {
            Device dev = mFramework.Device;

            Surface depth_buf = dev.DepthStencilSurface;
            Surface rt = dev.GetRenderTarget(0);

            dev.SetRenderTarget(0, mBuffer);
            dev.DepthStencilSurface = mDepthBuffer;
            dev.Clear(ClearFlags.Target,0,1.0f,0);

            mEffect.Render(mRenderables);

            dev.SetRenderTarget(0, rt);
            dev.DepthStencilSurface = depth_buf;
            
            rt.Dispose();
            depth_buf.Dispose();

            dev.GetRenderTargetData(mBuffer, mPickSurface);

            mRenderables.Clear();
        }

        public uint Pick(int x, int y)
        {
            if (mPickSurface == null)
                return 0;

            if ((x < 0) || (x >= mPickSurface.Description.Width))
                return 0;

            if ((y < 0) || (y >= mPickSurface.Description.Height))
                return 0;

            DataRectangle dr = mPickSurface.LockRectangle(new Rectangle(x, y, 1, 1), LockFlags.ReadOnly);
            
            uint data = dr.Data.Read<UInt32>();
            mPickSurface.UnlockRectangle();
            uint pick_id = CalculatePickID(data);
            return pick_id;
        }
        //public void SaveBuffer(Device dev)
        //{
        //    Surface tempo = Surface.CreateOffscreenPlain(dev, mBuffer.Description.Width, mBuffer.Description.Height, mBuffer.Description.Format, Pool.SystemMemory);
        //    dev.GetRenderTargetData(mBuffer, tempo);
        //    Surface.ToFile(tempo, "temp.bmp", ImageFileFormat.Bmp);
        //}
    }
}