//-----------------------------------------------------------------------------
// File: frames.cs
//
//-----------------------------------------------------------------------------

/*
alleg code for HardPoint:
.X -> mdl : done with "GroupGeoCallbackImpl::Execute" which is called by "mdlc -optimize"
.X -> cvh : done with xmunge -extract

in optimize geo: transform to stuff:
 * mat = transf matrix in X file frame
    Vector vecPosition = mat.Transform(Vector(0, 0, 0));
    Vector vecForward  = mat.TransformDirection(Vector(0, 0, -1));
    Vector vecUp       = mat.TransformDirection(Vector(0, 1,  0));

in xmunge:
 
   pframeChild->GetPosition(pframeTop, &pos);
   pframeChild->GetOrientation(pframeTop, &dir, &up);

 * frame -> transform matrix -> pos, fwd, up
  
so we have pos,fwd,up from each hardpoint but usage varies:
 
 * lights (*SS*): pos, MDL only
 * garageX?/capgarageX? : pos, forward, CVH only but appears in MDL
 * launchX : pos, forward : same as garage
 * missemt : pos (magazine & dispenser emit point) - used in CVH only
 * *wepemt* *turwepemt* : pos, forward
 * cockpt : pos
 * thrust : pos
 * weapon wepatt wepemt wepmnt : mount position - in MDL but seem not used?
 * trail, *rocket*, *smoke*: pos
 
 so "up" is never used althought stored in MDL files
 fwd is used in some cases althought always stored in MDL/CVH files

"pos" usage summary: always

"fwd" usage summary:
 * in MDL: never
 * in CVH: garage,launch and weapon positions (but see below)

"up" usage summary: never

 * code for weapon pos/fwd:
 
 m_scale = hull scale as defined in core
 pfd = framedata
 z = 0.0f, 0.0f, 1.0f
 
        m_positionWeapons[i] = pfd->position * m_scale;

        static const Vector up(0.0f, 1.0f, 0.0f);
        if (fabs(pfd->forward * up) < 0.9f)
            m_orientationWeapons[i].Set(pfd->forward, up);
        else
        {
            m_orientationWeapons[i].Set(pfd->forward, z);
        }
 
 * so if fabs(fwd*(0,1,0)) < 0.9f  the up is Y axis else it's Z axis.
 
 
 */

using System;
using System.Drawing;
using System.Collections;
using System.Collections.Generic;
using SlimDX;
using SlimDX.Direct3D9;
using Font = SlimDX.Direct3D9.Font;

using Microsoft.Samples.DirectX.UtilityToolkit;
using System.Text.RegularExpressions;

using Allegiance;
using Mindshifter;

namespace MDLEditor
{
    #region HardPoint
    public enum HardPointType
    {
        None, Light, Position, Launch, Garage, Effect
    };

    public struct HardPointPattern
    {
        public string pattern;
        public HardPointType type;
        public bool IsOriented;
    }

    public class HardPoint
    {
        public Vector3 Position { get; set; }
        public Vector3 Forward { get; set; }
        public Vector3 Up { get; set; } // unused atm but stored in MDL FrameData

        public HardPointType Type { get; set; }
        public bool Valid { get; set; }
        public bool IsOriented { get; set; }
        public string Name { get; set; }
        protected void Initialize(HardPointType type, string name)
        {
            Type = type;
            Name = name;
            IsOriented = false;
        }
        public HardPoint(HardPointType type, string name)
        {
            Initialize(type, name);
        }
        public HardPoint(HardPointType type, string name, bool valid)
        {
            Initialize(type, name);
            Valid = valid;
        }
        public void SetPFU(Matrix transform)
        {
            Vector4 v4 = Vector4.Transform(Vector4.UnitW,transform);
            Position = new Vector3(v4.X, v4.Y, v4.Z);
            System.Diagnostics.Debug.WriteLine("SetPFU " + v4.ToString());
            v4 = Vector4.Transform(-Vector4.UnitZ, transform);
            Forward = new Vector3(v4.X, v4.Y, v4.Z);

            v4 = Vector4.Transform(Vector4.UnitY, transform);
            Up = new Vector3(v4.X, v4.Y, v4.Z);
        }
        public Matrix GetTransform()
        {
            Matrix trans = Matrix.Translation(Position);
            if (IsOriented)
            {
                Matrix mfwd = new Matrix();
                mfwd.M31 = -Forward.X;
                mfwd.M32 = -Forward.Y;
                mfwd.M33 = -Forward.Z;
                trans *= mfwd;
            }
            return trans;
        }
        public Matrix Update(Matrix transform, Matrix worldMatrix)
        {
            return transform * worldMatrix;
        }
        public const string FrmXPreffix = "frm-"; 

        public static string SubFrmStringRegex(string sub)
        {
            return "^"+FrmXPreffix+"(?<name>.*(" + sub + ").*)$";
        }
        public static string SubStringRegex(string sub)
        {
            return "(?<name>.*(" + sub + ").*)$";
        }
        public static HardPointPattern[] AllegPatterns()
        {
            return new HardPointPattern[] {
                new HardPointPattern{ pattern = SubFrmStringRegex("wepemt"),                                   type = HardPointType.Position, IsOriented = true},
                new HardPointPattern{ pattern = SubFrmStringRegex("turwepemt"),                                type = HardPointType.Position, IsOriented = true},
                new HardPointPattern{ pattern = SubFrmStringRegex("missemt"),                                  type = HardPointType.Position, IsOriented = false},
                new HardPointPattern{ pattern = "^"+FrmXPreffix+"(?<name>cockpt)$",                                    type = HardPointType.Position, IsOriented = false},
                new HardPointPattern{ pattern = "^"+FrmXPreffix+"(?<name>launch[1-9])$",                               type = HardPointType.Launch, IsOriented = true},
                new HardPointPattern{ pattern = "^"+FrmXPreffix+"(?<name>capgarage)(?<num>[1-9])(?<letter>[A-E])$",    type = HardPointType.Garage, IsOriented = true},
                new HardPointPattern{ pattern = "^"+FrmXPreffix+"(?<name>garage)(?<num>[1-9])(?<letter>[A-E])$",       type = HardPointType.Garage, IsOriented = true},
                new HardPointPattern{ pattern = SubFrmStringRegex("SS"),                                       type = HardPointType.Light, IsOriented = false},
                new HardPointPattern{ pattern = SubFrmStringRegex("thrust"),                                   type = HardPointType.Effect, IsOriented = false},
                new HardPointPattern{ pattern = SubFrmStringRegex("trail"),                                    type = HardPointType.Effect, IsOriented = false},
                new HardPointPattern{ pattern = SubFrmStringRegex("smoke"),                                    type = HardPointType.Effect, IsOriented = false},
                new HardPointPattern{ pattern = SubFrmStringRegex("rocket"),                                   type = HardPointType.Effect, IsOriented = false}
            };
        }
        public static HardPointPattern[] AllegCVHPatterns()
        {
            return new HardPointPattern[] {
                new HardPointPattern{ pattern = SubStringRegex("wepemt"),                                   type = HardPointType.Position, IsOriented = true},
                new HardPointPattern{ pattern = SubStringRegex("turwepemt"),                                type = HardPointType.Position, IsOriented = true},
                new HardPointPattern{ pattern = SubStringRegex("missemt"),                                  type = HardPointType.Position, IsOriented = true},
                new HardPointPattern{ pattern = "^(?<name>cockpt)$",                                    type = HardPointType.Position, IsOriented = false},
                new HardPointPattern{ pattern = "^(?<name>launch[1-9])$",                               type = HardPointType.Launch, IsOriented = true},
                new HardPointPattern{ pattern = "^(?<name>capgarage)(?<num>[1-9])(?<letter>[A-E])$",    type = HardPointType.Garage, IsOriented = true},
                new HardPointPattern{ pattern = "^(?<name>garage)(?<num>[1-9])(?<letter>[A-E])$",       type = HardPointType.Garage, IsOriented = true},
            };
        }

        public static HardPoint GetHardPoint(string name)
        {
            foreach (HardPointPattern pat in AllegPatterns())
            {
                Regex regex = new Regex(pat.pattern);
                Match m = regex.Match(name);
                if (m.Success)
                {
                    HardPoint hp;
                    switch (pat.type)
                    {
                        case HardPointType.Garage:
                        {
                            hp = new HardPointGarage(m.Groups["name"].Value, m.Groups["num"].Value, m.Groups["letter"].Value);
                            break;
                        }
                        case HardPointType.Light:
                        {
                            hp = new HardPointLight(m.Groups["name"].Value);
                            break;
                        }
                        default:
                        {
                            hp = new HardPoint(pat.type, m.Groups["name"].Value, true);
                            break;
                        }
                    }
                    hp.IsOriented = pat.IsOriented;
                    return hp;
                }
            }
            return new HardPoint(HardPointType.None, name, true);
        }
        public static HardPoint GetCVHHardPoint(string name)
        {
            foreach (HardPointPattern pat in AllegCVHPatterns())
            {
                Regex regex = new Regex(pat.pattern);
                Match m = regex.Match(name);
                if (m.Success)
                {
                    HardPoint hp;
                    if (pat.type == HardPointType.Garage)
                       hp = new HardPointGarage(m.Groups["name"].Value, m.Groups["num"].Value, m.Groups["letter"].Value);
                    else 
                        hp = new HardPoint(pat.type, m.Groups["name"].Value, true);
                    hp.IsOriented = pat.IsOriented;
                    return hp;
                }
            }
            // should we throw an execption or ignore the HP ?
            return new HardPoint(HardPointType.None, name, true);
        }
    }
    public class HardPointLight : HardPoint
    {
        public MDLLight LightData { get ; set; }
        public HardPointLight(string name) : base(HardPointType.Light,name)
        {
            LightData = new MDLLight();
        }
    }

    public class HardPointGarage : HardPoint
    {
        public string Root   { get; set; } // garage or capgarage
        public string Number { get; set; } // 1 to 9
        public string Letter { get; set; } // A to E
        public AnimationFrame[] Planes { get; set; } // 
        public HardPointGarage(string root, string num, string letter)
            : base(HardPointType.Garage, root + num)
        {
            Root = root;
            Number = num;
            Letter = letter;
        }
    }

    #endregion

    public enum RenderStyle
    {
        Solid, Wireframe, Hide
    }

    #region AnimationFrame Class
    /// <summary>
    /// The frame that will hold mesh animation 
    /// </summary>
    public class AnimationFrame : Frame, IBoundedTransform
    {
        private BoundingSphere boundingsphere;
        private BoundingBox boundingsbox;
        // Store the combined transformation matrix
        private Matrix combined = Matrix.Identity;
        private Matrix inherited = Matrix.Identity;
        private Matrix parentTrans = Matrix.Identity;
        /// <summary>The combined transformation matrix</summary>
        public Matrix CombinedTransformationMatrix
        {
            get { return combined; } set { combined = value; }
        }
        public Matrix InheritedTransformationMatrix
        {
            get { return inherited; }
            set { inherited = value; }
        }
        public RenderStyle Style { get; set; }
        public HardPoint HP { get; set; }
        public void ComputeBoundingSphere()
        {
            boundingsphere = Frame.CalculateBoundingSphere(this);
            boundingsbox = BoundingBox.FromSphere(boundingSphere);
        }
        public BoundingSphere boundingSphere
        {
            get { return boundingsphere; } 
        }
        private bool isconvexhull = false;
        public bool IsConvexHull { get { return isconvexhull; } set { isconvexhull = value; } }

        public void Update(Matrix parentMatrix, Matrix worldMatrix)
        {
            parentTrans = parentMatrix;
            if (HP.Type == HardPointType.None)
            {
                CombinedTransformationMatrix = TransformationMatrix * worldMatrix;
                InheritedTransformationMatrix = TransformationMatrix * parentMatrix;
            }
            else
            {
                CombinedTransformationMatrix = HP.Update(TransformationMatrix,worldMatrix);
                InheritedTransformationMatrix = HP.Update(TransformationMatrix,parentMatrix);
            }
        }
        // ITransform members
        private Vector3 tmTrans;
        private Quaternion tmRot;
        private Vector3 tmScale;
        private void UpdateTM(bool decompose)
        {
            if (decompose)
                InheritedTransformationMatrix.Decompose(out tmScale, out tmRot, out tmTrans);
            else
                // todo
                TransformationMatrix = Matrix.Scaling(tmScale) * Matrix.RotationQuaternion(tmRot) * Matrix.Translation(tmTrans);
        }
        public Vector3 Translation { get { UpdateTM(true); return tmTrans; } set { tmTrans = value; UpdateTM(false); } }
        public Quaternion Rotation { get { UpdateTM(true); return tmRot; } set { tmRot = value; UpdateTM(false); } }
        public Vector3 Scale { get { UpdateTM(true); return tmScale; } set { tmScale = value; UpdateTM(false); } }
        // IBoundedTransform members
        public BoundingBox Bounds { get { return boundingsbox;} }
    } // AnimatedFrame
    #endregion
    #region Derived Mesh Container
    /// <summary>
    /// The mesh container class that will hold the animation data
    /// </summary>
    public class AnimationMeshContainer : MeshContainer
    {
        // Array data
        private Texture[] meshTextures = null;
        private BoneCombination[] bones;
        private Matrix[] offsetMatrices;
        private AnimationFrame[] frameMatrices;

        // Instance data
        private int numAttributes = 0;
        private int numInfluences = 0;
        private int numPalette = 0;

        // Public properties

        /// <summary>Retrieve the textures used for this container</summary>
        public Texture[] GetTextures() { return meshTextures; }
        /// <summary>Set the textures used for this container</summary>
        public void SetTextures(Texture[] textures) { meshTextures = textures; }
        
        /// <summary>Retrieve the bone combinations used for this container</summary>
        public BoneCombination[] GetBones() { return bones; }
        /// <summary>Set the bone combinations used for this container</summary>
        public void SetBones(BoneCombination[] b) { bones = b; }
        
        /// <summary>Retrieve the animation frames used for this container</summary>
        public AnimationFrame[] GetFrames() { return frameMatrices; }
        /// <summary>Set the animation frames used for this container</summary>
        public void SetFrames(AnimationFrame[] frames) { frameMatrices = frames; }
        
        /// <summary>Retrieve the offset matrices used for this container</summary>
        public Matrix[] GetOffsetMatrices() { return offsetMatrices; }
        /// <summary>Set the offset matrices used for this container</summary>
        public void SetOffsetMatrices(Matrix[] matrices) { offsetMatrices = matrices; }
        
        /// <summary>Total number of attributes this mesh container contains</summary>
        public int NumberAttributes { get { return numAttributes; } set { numAttributes = value; } }
        /// <summary>Total number of influences this mesh container contains</summary>
        public int NumberInfluences { get { return numInfluences; } set { numInfluences = value; } }
        /// <summary>Total number of palette entries this mesh container contains</summary>
        public int NumberPaletteEntries { get { return numPalette; } set { numPalette = value; } }
    }
    #endregion
    #region Animation Allocation Hierarchy
    public interface IEAllocateHierarchy : IAllocateHierarchy
    {
        Frame CreateFrame(string name, HardPoint hp);
    }
    /// <summary>
    /// AllocateHierarchy derived class
    /// </summary>
    public class AnimationAllocation : IEAllocateHierarchy
    {
        AMTForm parent = null;
        /// <summary>Create new instance of this class</summary>
        public AnimationAllocation(AMTForm p) { parent = p; }

        protected AnimationFrame CreateFrameInternal(string name)
        {
            AnimationFrame frame = new AnimationFrame();
            frame.Name = name;
            frame.TransformationMatrix = Matrix.Identity;
            frame.CombinedTransformationMatrix = Matrix.Identity;
            frame.Style = RenderStyle.Solid;
            return frame;
        }
        /// <summary>Create a new frame, parse name for HP</summary>
        public Frame CreateFrame(string name)
        {
            AnimationFrame frame = CreateFrameInternal(name);
            frame.HP = HardPoint.GetHardPoint(name);
            return frame;
        }
        /// <summary>Create a new frame, given an HP</summary>
        public Frame CreateFrame(string name, HardPoint hp)
        {
            AnimationFrame frame = CreateFrameInternal(name);
            frame.HP = hp;
            return frame;
        }
        /// <summary>Create a new mesh container</summary>
        public MeshContainer CreateMeshContainer(string name, 
            MeshData meshData, ExtendedMaterial[] materials, 
            EffectInstance[] effectInstances, int[] adjacency, 
            SkinInfo skinInfo)
        {
            // We only handle meshes here
            if (meshData.Mesh == null)
                throw new ArgumentException();

            // We must have a vertex format mesh
            if (meshData.Mesh.VertexFormat == VertexFormat.None)
                throw new ArgumentException();

            AnimationMeshContainer mesh = new AnimationMeshContainer();

            mesh.Name = name;
            int numFaces = meshData.Mesh.FaceCount;
            Device dev = meshData.Mesh.Device;
            
            // Make sure there are normals
            if ((meshData.Mesh.VertexFormat & VertexFormat.Normal) == 0)
            {
                // Clone the mesh
                Mesh tempMesh = meshData.Mesh.Clone(dev, meshData.Mesh.CreationOptions, 
                    meshData.Mesh.VertexFormat | VertexFormat.Normal);

                // Destroy current mesh, use the new one
                meshData.Mesh.Dispose();
                meshData = new MeshData(tempMesh);
                meshData.Mesh.ComputeNormals();
            }

            // Store the materials
            if (adjacency != null)
                mesh.SetAdjacency(adjacency);

            if (materials != null)
            {
                mesh.SetMaterials(materials);
                Texture[] meshTextures = new Texture[materials.Length];

                // Create any textures
                for (int i = 0; i < materials.Length; i++)
                {
                    if (materials[i].TextureFileName != null)
                        if (materials[i].TextureFileName != "")
                        {
                            try
                            {
                                meshTextures[i] = MDLResourceCache.GetGlobalInstance().CreateTextureFromFile(
                                    dev, materials[i].TextureFileName, true);
                            }
                            catch
                            {
                                meshTextures[i] = parent.defaultTexture;
                            }
                        }
                }
                mesh.SetTextures(meshTextures);
            }
            mesh.MeshData = meshData;

            // If there is skinning info, save any required data
            if (skinInfo != null)
            {
                mesh.SkinInfo = skinInfo;
                int numBones = skinInfo.BoneCount;
                Matrix[] offsetMatrices = new Matrix[numBones];

                for (int i = 0; i < numBones; i++)
                    offsetMatrices[i] = skinInfo.GetBoneOffsetMatrix(i);

                mesh.SetOffsetMatrices(offsetMatrices);

                parent.GenerateSkinnedMesh(mesh);
            }

            return mesh;
        }
        public void DestroyFrame(Frame frame)
        {
            frame.Dispose();
        }
        public void DestroyMeshContainer(MeshContainer container)
        {
            container.Dispose();
        }
    }
    #endregion
    #region Helper class
    public class AMTHelper
    { 
        /// <summary>
        /// advance anim to 1st keyed frame
        /// assumes there is only one AnimationSet 
        /// </summary>
        /// <param name="controller"></param>
        /// <param name="key"></param>
        public static void MoveAnimToKey(AnimationController controller, int key)
        {
            AnimationSet aset = controller.GetAnimationSet(0); // assumption here
            double time = aset.GetPeriodicPosition(key);
            controller.ResetTime(); 
            controller.AdvanceTime(time,null);
        }
        /// <summary>
        /// merge garage frames into 1
        /// </summary>
        /// <param name="rootframe"></param>
        public static void MergeGarageFrames(AnimationFrame rootframe, AnimationFrame frame, bool bFrmPrefix)
        {
            if (frame.HP.Type == HardPointType.Garage)
            {
                HardPointGarage gar = frame.HP as HardPointGarage;
                if (gar.Letter == "A")
                {
                    string frmprefix = bFrmPrefix ? HardPoint.FrmXPreffix : "";
                    List<AnimationFrame> planes = new List<AnimationFrame>(6);
                    planes.Add(frame);
                    foreach (string letter in new string[] { "B", "C", "D", "E", "F" })
                    {
                        AnimationFrame plane = rootframe.FindChild(frmprefix + gar.Root + gar.Number + letter) as AnimationFrame;
                        if (plane != null)
                            planes.Add(plane);
                    }
                    gar.Valid = true;
                }
            }
            // Next do any siblings this frame may contain
            if (frame.Sibling != null)
            {
                MergeGarageFrames(rootframe, frame.Sibling as AnimationFrame, bFrmPrefix);
            }
            // Finally do the children of this frame
            if (frame.FirstChild != null)
            {
                MergeGarageFrames(rootframe, frame.FirstChild as AnimationFrame, bFrmPrefix);
            }
        }
        public static AnimationFrame CVHtoFrame(IEAllocateHierarchy alloc, CVH cvh, Device device)
        {
            if (cvh == null) return null;
            AnimationFrame rootcvh = alloc.CreateFrame("CVH") as AnimationFrame;
            rootcvh.IsConvexHull = true;
            int n=0;
            Frame curframe = rootcvh;
            foreach (ConvexHull hull in cvh.Hulls)
            {
                AnimationFrame hullframe = alloc.CreateFrame("CVH-" + n.ToString()) as AnimationFrame;
                hullframe.IsConvexHull = true;
                ExtendedMaterial mat = new ExtendedMaterial();
                MeshData meshdata = new MeshData(ConvexHulltoMesh(hull,device));
                hullframe.MeshContainer = alloc.CreateMeshContainer("hull" + n.ToString(),
                    meshdata, new ExtendedMaterial[] {mat}, null, null, null);
                if (n == 0)
                    curframe.FirstChild = hullframe;
                else
                    curframe.Sibling = hullframe;
                curframe = hullframe;
                n++;
            }
            foreach (CVHFrame frame in cvh.Frames)
            {
                HardPoint hp = HardPoint.GetCVHHardPoint(frame.Name);
                if (hp.Type == HardPointType.None) continue;

                AnimationFrame pf = alloc.CreateFrame(frame.Name,hp) as AnimationFrame;
                hp.Position = frame.Position;
                hp.Forward = frame.Forward;
                pf.TransformationMatrix = hp.GetTransform();
                rootcvh.AppendChild(pf);
            }
            return rootcvh;
        }
        public static Mesh ConvexHulltoMesh(ConvexHull hull, Device device)
        {
            hull.GenMesh();
            Int16[] faces = hull.Mesh.Faces;
            Mesh mesh = new Mesh(device, faces.Length/3, hull.Mesh.Vertices.Length, MeshFlags.SystemMemory, VertexFormat.Position);
            // vertices
            {
                DataStream ds = mesh.LockVertexBuffer(LockFlags.None);
                foreach (Vector3 v in hull.Mesh.Vertices)
                {
                    ds.Write<Vector3>(v);
                }
                mesh.UnlockVertexBuffer();
            }
            // indices
            {
                DataStream ds = mesh.LockIndexBuffer(LockFlags.None);
                ds.WriteRange<Int16>(faces);
                mesh.UnlockIndexBuffer();
            }

            return mesh;
        }
        public static AnimationFrame MDLtoFrame(IEAllocateHierarchy alloc, MDLFile mdl, Device device)
        {
            if (mdl == null) return null;
            MDLObject rootobject = mdl.RootObject;
            if (mdl.RootObject.type == MDLType.mdl_lod)
            {
                rootobject = mdl.RootObject.childrens[0];
            }
            ExtendedMaterial[] mats = new ExtendedMaterial[mdl.Textures.Length];
            int i = 0;
            Material dmat = new Material(); dmat.Diffuse = Color.White;
            foreach (string texname in mdl.Textures)
            {
                mats[i].MaterialD3D = dmat;
                mats[i].TextureFileName = texname + ".mdl";
                i++;
            }
            AnimationFrame frame = MDLObjecttoFrame(alloc, rootobject, device, 0, 0, mats);
            i = 0;
            foreach (MDLLight l in mdl.Lights)
            {
                HardPointLight ld = new HardPointLight("SS_" + i.ToString());
                ld.LightData = l;
                ld.Position = l.position;
                AnimationFrame lf = alloc.CreateFrame(ld.Name,ld) as AnimationFrame;
                lf.TransformationMatrix = ld.GetTransform();
                frame.AppendChild(lf);
                i++;
            }
            foreach (MDLFrameData f in mdl.FrameDatas)
            {
                HardPoint hp = HardPoint.GetCVHHardPoint(f.name);
                if (hp.Type != HardPointType.Garage)
                    hp = new HardPoint(HardPointType.Effect, f.name);
                hp.Position = f.vecPosition;
                hp.Forward = f.vecForward;
                AnimationFrame ff = alloc.CreateFrame(f.name, hp) as AnimationFrame;
                ff.TransformationMatrix = hp.GetTransform();
                frame.AppendChild(ff);
            }
            return frame;
        }
        public static AnimationFrame MDLObjecttoFrame(IEAllocateHierarchy alloc, MDLObject mdl, Device device, int level, int num, ExtendedMaterial[] gmats)
        {
            string name = (level == 0) ? "MDLRoot" : "MDL"+level.ToString()+"_"+num.ToString();
            AnimationFrame frame = alloc.CreateFrame(name) as AnimationFrame;
            if (mdl.type == MDLType.mdl_group)
            {
                int n = 1;
                foreach (MDLObject child in mdl.childrens)
                {
                    frame.AppendChild(MDLObjecttoFrame(alloc,child,device,level+1, n++, gmats));
                }
            }
            if (mdl.type == MDLType.mdl_mesh)
            {
                ExtendedMaterial[] mats;
                if (mdl.textidx != -1)
                {
                    mats = new ExtendedMaterial[1];
                    mats[0] = gmats[mdl.textidx];
                }
                else
                    mats = null;
                MeshData meshdata = new MeshData(MDLMeshtoMesh(mdl.mesh, device));
                frame.MeshContainer = alloc.CreateMeshContainer(name+"_mesh",
                    meshdata, mats , null, null, null);
            }
            return frame;
        }
        public static Mesh MDLMeshtoMesh(MDLMesh mdlmesh, Device device)
        {
            Mesh mesh = new Mesh(device, mdlmesh.nfaces, mdlmesh.nvertex, MeshFlags.SystemMemory, MDLVertexFormat);
            DataStream ds = mesh.LockVertexBuffer(LockFlags.None);
            foreach (MDLVertice v in mdlmesh.vertices)
            {
                ds.Write<float>(v.position.X);ds.Write<float>(v.position.Y);ds.Write<float>(v.position.Z);
                ds.Write<float>(v.normal.X); ds.Write<float>(v.normal.Y); ds.Write<float>(v.normal.Z);
                ds.Write<float>(v.uv.X); ds.Write<float>(v.uv.Y);
            }
            mesh.UnlockVertexBuffer();
            ds = mesh.LockIndexBuffer(LockFlags.None);
            ds.WriteRange<ushort>(mdlmesh.backfaces);
            mesh.UnlockIndexBuffer();
            return mesh;
        }
        public const VertexFormat MDLVertexFormat = VertexFormat.Position | VertexFormat.Normal | VertexFormat.Texture1;
    }
    #endregion
}
